// all the code and simulation infrastructure can be obtained from http://csg.csail.mit.edu/6.S078/6_S078_2012_www/handouts.html
// cp0_tohost, cp0_fromhost, cp0_statsEn, num_cycles, num_inst are all debugging state, and are not present in the code presented in the lecture slides
// mkBTB, mkDCache, mkICache, mkMemory, mkEHRBypassRFile are all synthesized at compile time, and help to reduce the compilation time when re-compiling after a minor change

import ProcTypes::*;
import MemTypes::*;
import RFile::*;
import Decode::*;
import Exec::*;
import EHRPipeReg::*;
import EHR::*;
import Btb::*;
import ICache::*;
import DCache::*;
import Memory::*;
import Scoreboard::*;
import Connectable::*;

interface Proc;
    method Bit#(32) cpuToHost(Bit#(2) req);
    method Action hostToCpu(Bit#(32) startpc);
endinterface

(* synthesize *)
module mkProc(Proc);
    EHR3#(Addr)             pc <- mkEHR3U;
    RFile                   rf <- mkEHRBypassRFile;
    Scoreboard#(4)          sb <- mkScoreboard;
    EHR3#(Bool)          epoch <- mkEHR3(False);
    NextAddressPredictor bpred <- mkBTB;
    ICache              iCache <- mkICache;
    DCache              dCache <- mkDCache;
    Memory                 mem <- mkMemory;

    EHRBypassReg#(TypeFetch2Fetch)         fr <- mkEHRBypassReg;
    EHRPipeReg#(TypeFetch2Decode)          ir <- mkEHRPipeReg;
    EHRPipeReg#(TypeDecode2Execute)       itr <- mkEHRPipeReg;
    EHRPipeReg#(TypeExecute2Commit)        er <- mkEHRPipeReg;
    EHRBypassReg#(TypeExecute2Commit)      mr <- mkEHRBypassReg;
    EHRPipeReg#(TypeExecute2Commit)        cr <- mkEHRPipeReg;
    
    EHR3#(Data)      cp0_tohost <- mkEHR3(0);
    Reg#(Bool)     cp0_fromhost <- mkReg(False);
    EHR3#(Bool)     cp0_statsEn <- mkEHR3(False);
    Reg#(Stat)       num_cycles <- mkReg(0);
    EHR3#(Stat)        num_inst <- mkEHR3(0);

    rule doFetch1(cp0_fromhost && fr.notFull);
        $display("Proc: fetch1 %h", pc.r1);

        iCache.req(MemReq{op:Ld, addr:pc.r1, data:?});

        let ppc = bpred.prediction(pc.r1);

        fr.enq(TypeFetch2Fetch{pc:pc.r1, ppc:ppc, epoch:epoch.r1});
        pc.w1(ppc);
    endrule

    rule doFetch2(fr.notEmpty && ir.notFull);
        let frpc = fr.first.pc;
        let frppc = fr.first.ppc;
        let frepoch = fr.first.epoch;

        $display("Proc: fetch2 %h", frpc);

        let inst = iCache.resp;

        ir.enq(TypeFetch2Decode{pc:frpc, ppc:frppc, epoch:frepoch, inst:inst});
        
        iCache.respDeq;
        fr.deq;
    endrule

    rule doDecode(ir.notEmpty && itr.notFull);
        let irpc = ir.first.pc;
        let irppc = ir.first.ppc;
        let irepoch = ir.first.epoch;
        let inst = ir.first.inst;

        let dInst = decode(inst, cp0_statsEn.r1, cp0_fromhost, cp0_tohost.r1);

        let stall = sb.search(dInst.rSrc1, dInst.rSrc2);
        if(!stall) begin
            $display("Proc: decode %h", irpc);

            let rVal1 = rf.rd1(unJust(dInst.rSrc1));
            let rVal2 = rf.rd2(unJust(dInst.rSrc2));

            itr.enq(TypeDecode2Execute{pc:irpc, ppc:irppc, epoch:irepoch, inst:inst, dInst:dInst, rVal1:rVal1, rVal2:rVal2});
            sb.insert(dInst.rDst);
            ir.deq;
        end
    endrule

    rule doExecute(itr.notEmpty && er.notFull);
        let itrpc = itr.first.pc;
        let itrppc = itr.first.ppc;
        let inst = itr.first.inst;
        let dInst = itr.first.dInst;
        let rVal1 = itr.first.rVal1;
        let rVal2 = itr.first.rVal2;

        if(itr.first.epoch==epoch.r0) begin
            let eInst = exec(dInst, rVal1, rVal2, itrpc, itrppc, ?);
            Instr i = unpack(inst);
            traceFull("Proc", "exec", i);

            er.enq(TypeExecute2Commit{pc:itrpc, eInst:eInst});

            if(eInst.misprediction) begin
                let npc = eInst.brTaken ? eInst.addr : itrpc+4;
                pc.w0(npc);
                epoch.w0(!epoch.r0);
                bpred.update(itrpc, npc);
            end

            if(eInst.iType==Mtc0) begin
                if(eInst.wrMtc0)
                    cp0_statsEn.w0(unpack(truncate(rVal2)));
                else
                    cp0_tohost.w0(rVal2);
            end
            if(cp0_statsEn.r0) num_inst.w0(num_inst.r0 + 1);
        end
        else begin
            ExecInst eInst = ?;
            eInst.iType = Nop;
            er.enq(TypeExecute2Commit{pc:itrpc, eInst:eInst});
        end

        itr.deq;
    endrule

    rule doMemory1(er.notEmpty && mr.notFull);
        let erpc = er.first.pc;
        let eInst = er.first.eInst;

        $display("Proc: memory1 %h", erpc);

        if(memType(eInst.iType))
            dCache.req(MemReq{op:eInst.iType==Ld ? Ld : St, addr:eInst.addr, data:eInst.data});

        mr.enq(TypeExecute2Commit{pc:erpc, eInst:eInst});

        er.deq;
    endrule

    rule doMemory2(mr.notEmpty && cr.notFull);
        let mrpc = mr.first.pc;
        let eInst = mr.first.eInst;

        $display("Proc: memory2 %h", mrpc);

        if(eInst.iType==Ld) begin
            eInst.data = dCache.resp;
            dCache.respDeq;
        end

        cr.enq(TypeExecute2Commit{pc:mrpc, eInst:eInst});

        mr.deq;
    endrule

    rule doCommit(cr.notEmpty);
        let crpc = cr.first.pc;
        let eInst = cr.first.eInst;

        $display("Proc: commit %h", crpc);

        if(regWriteType(eInst.iType))
            rf.wr(unJust(eInst.rDst), eInst.data);
        
        sb.remove;
        cr.deq;
    endrule

    rule doCyc(cp0_statsEn.r2);
        num_cycles <= num_cycles+1;
    endrule

    mkConnection(mem.iReq, iCache.mReq);
    mkConnection(mem.iResp, iCache.mResp);
    mkConnection(mem.dReq, dCache.mReq);
    mkConnection(mem.dResp, dCache.mResp);

    method Bit#(32) cpuToHost(Bit#(2) req);
        return (case (req)
            0: cp0_tohost.r2;
            1: pc.r2;
            2: num_inst.r2;
            3: num_cycles;
        endcase);
    endmethod
    method Action hostToCpu(Bit#(32) startpc);
        cp0_fromhost <= True;
        pc.w2(startpc);
    endmethod
endmodule

